#
# $Copyright: Copyright (c) 2016 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#
# $RCSfile: SVCCopyServicesAgent.pm,v $ $Revision: 1.2.4.28 $
#

package SVCCopyServicesAgent;
use ArrayRep;
@ISA = ("ArrayRep");

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default);
use ag_i18n_inc;
my $vcs_home = $ag_i18n_inc::VCSHOME;

@EXPORT = qw($ACTION_DONOTHING_ISONLINE,
	$ACTION_DONOTHING_ISOFFLINE, $ACTION_SWITCHRC,
	$ACTION_STOPTAKEOVER, $ACTION_DISCTAKEOVER, $ACTION_COPYTAKEOVER,
	$STATE_IDLING, $STATE_IDLING_DISCONNECTED, $STATE_CONSISTENT_STOPPED,
	$STATE_CONSISTENT_DISCONNECTED, $MASTER, $AUX,
	$STOP_TAKEOVER, $DISCONNECT_TAKEOVER);

$VERSION = '1.00';

require 5.000;

our ($OSTYPE);

my $SVCCONFIG_CMD = "svcconfig";
my $SVCVERSION = "-ver";
my $SVCINFO_CMD = "svcinfo";
my $SVCTASK_CMD = "svctask";
my $LSCLUSTER = "lscluster";
my $LSCLUSTERIP = "lsclusterip";
my $LSRC = "lsrc";
my $LSREL = "lsrcrelationship";
my $LSVDISK = "lsvdisk";
my $STARTRC = "startrc";
my $SWITCHRC = "switchrc";
my $STOPRC = "stoprc";
my $CHRC = "chrc";
my $ACCESS_OPT = "-access";
my $MASTER_CLUSTER_ID = "master_cluster_id";
my $AUX_CLUSTER_ID = "aux_cluster_id";
our $MASTER = "master";
our $AUX = "aux";
our $LOCAL = "local";
my $COPYTYPE_METROMIRROR = "metro";
my $COPYTYPE_GLOBALMIRROR = "global";
our $STOP_TAKEOVER = 0;
our $DISCONNECT_TAKEOVER = 1;
our $COPY_TAKEOVER = 2;
my $COPY_TYPE = "copy_type";
my $VXDCTL = "/usr/sbin/vxdctl";
my $VXDISK = "/usr/sbin/vxdisk";

# This action signifies that we only need to create the lockfile and
# exit the online ep
our $ACTION_DONOTHING_ISONLINE = 0;

# This action signifies that we cant online the resource. So exit.
our $ACTION_DONOTHING_ISOFFLINE = 1;

# This action signifies that we need to invoke the switchrc cmd to 
# take over as the Primary
our $ACTION_SWITCHRC = 2;

# This action signifies that we do a takeover if the replication is 
# in a Stopped state - even if there is currently no read/write access
our $ACTION_STOPTAKEOVER = 3;

# This action signifies that we do a takeover if the two SVC clusters 
# participating in the replication are Disconnected - even if there is 
# currently no read/write access
our $ACTION_DISCTAKEOVER = 4;

# This action signifies that we do a takeover if the replication is 
# in a consistent_copying state - even if there is currently no read/write
# access
our $ACTION_COPYTAKEOVER = 5;

# Valid states for a MM/GM relationship or a consistency group
my @VALID_STATES =	("inconsistent_stopped",
			"inconsistent_copying",
			"inconsistent_disconnected",
			"consistent_stopped",
			"consistent_synchronized",
			"consistent_disconnected",
			"idling",
			"idling_disconnected",
			"empty",
			"consistent_copying"
			);
our $STATE_INCONSISTENT_STOPPED = "inconsistent_stopped";
our $STATE_INCONSISTENT_COPYING = "inconsistent_copying";
our $STATE_INCONSISTENT_DISCONNECTED = "inconsistent_disconnected";
our $STATE_CONSISTENT_STOPPED = "consistent_stopped";
our $STATE_CONSISTENT_SYNCHRONIZED = "consistent_synchronized";
our $STATE_CONSISTENT_DISCONNECTED = "consistent_disconnected";
our $STATE_IDLING = "idling";
our $STATE_IDLING_DISCONNECTED = "idling_disconnected";
our $STATE_EMPTY = "empty";
our $STATE_CONSISTENT_COPYING = "consistent_copying";

# Values of the "sync" attribute that indicate that the data is in sync
# between the master and aux clusters.
my @INSYNC_STATUS = ( "", "in_sync" );

%default = (
	resname => '',

	# Name of the service group the SVC resource belongs to
	sgname => '',

	# Name of the metro/global mirror relationship or consistency
	# group to be managed
	name	=> '',

	# Type of relationship to manage: Consistency Group or MM/GM
	# relationship; 0 by default
	is_consistency_group => '',

	# All cmds need to have the suffix "relationship" when being
	# invoked for a MM/GM relationship:
	# E.g. svcinfo lsrcrelationship
	# For a consistency group, the cmd suffix will be "consistgrp"
	# E.g. svcinfo lsrcconsistgrp
	cmd_suffix => '',

	# Login to the SVC cluster as: $ssh_bin -i $ssh_identity_file -l
	# $svc_username $svc_cluster_ip <cmd to be run>

	# Path to SSH binary
	ssh_binary	=> '',

	# Path to SSH binary
	ssh_binary_w_quotes	=> '',

	# SSH Private Key File - needed to talk with the SVC cluster
	ssh_identity_file => '',

	# SSH Private Key File - needed to talk with the SVC cluster
	ssh_identity_file_w_quotes => '',

	# SVC Cluster IP
	svc_cluster_ip => '',

	# SVC user name to login as
	svc_username => '',

	# For future use - whether global or metro
	copy_type	=> '',

	# Could be one of: InconsistentStopped, InconsistentCopying,
	# ConsistentStopped, ConsistentSync, InconsistentDisconnected,
	# ConsistentDisconnected, Idling, IdlingDisconnected
	state 	=> '',

	# Whether master or aux
	role 	=> '',

	# Start time - of the entry point
	start_time => '',

	# Number of seconds to extend action tmo by
	entrypoint_timeout_ext => '',

	# Has tmo been extended?
	is_timeout_extended => '',

	# Resource type the resource belongs to
	resource_type => '',

	# Set to 1 if the service grp containing the resource
	# is a parallel service grp; 0 otherwise
	rac => '',

	# Cycling mode file
	cycling_mode_file => '',

	# SVC version
	svc_version => '',
);

sub new {
	my ($proto, $resname, $name, $is_cg, $ssh_bin, $ssh_identity_file, $svc_cluster_ip, $svc_username) = @_;
	my $class = ref($proto) || $proto;
	my $self = { %default };
	$self->{'start_time'} = time();
	$self->{'resource_type'} = "SVCCopyServices";
	$self->{'resname'} = $resname;
	$self->{'name'} = $name;
	$self->{'is_consistency_group'} = $is_cg;
	$self->{'copy_type'} = "";

	$self->{'ssh_binary'} = $ssh_bin;
	$self->{'ssh_identity_file'} = $ssh_identity_file;
	$self->{'svc_cluster_ip'} = $svc_cluster_ip;
	$self->{'svc_username'} = $svc_username;
	$self->{'cmd_suffix'} = "relationship";

	$self->{'rac'} = 0;
	
	my ($sg) = split('\n', `$ag_i18n_inc::HARES -value $resname Group`);
	$self->{'sgname'} = $sg;

	my ($pll) = split('\n', `$ag_i18n_inc::HAGRP -value $sg Parallel`);
	if ($pll == 1) {
		$self->{'rac'} = 1;
	}

	if ( $^O =~ /^MSWin/i) {
		require Win32;
		Win32->import;

		$self->{'ssh_binary'} = Win32::GetShortPathName($ssh_bin);
		$self->{'ssh_identity_file'} = Win32::GetShortPathName($ssh_identity_file);
		$self->{'ssh_binary_w_quotes'} = "\"". Win32::GetShortPathName($ssh_bin). "\"";
		$self->{'ssh_identity_file_w_quotes'} = "\"". Win32::GetShortPathName($ssh_identity_file). "\"";

		my $vmpath;
		$vmpath = $ENV{"VMPATH"};
		$vmpath = "\"". Win32::GetShortPathName($vmpath). "\"";
		$VXDCTL = "$vmpath" . "vxdctl";
		$VXDISK = "$vmpath" . "vxdisk";
		$self->{'rac'} = 0;    
  	} else {
		$self->{'ssh_binary_w_quotes'} = "\"". $ssh_bin. "\"";
		$self->{'ssh_identity_file_w_quotes'} = "\"". $ssh_identity_file. "\"";
	}

	if ($is_cg == 1) {
		# Change the cmd_suffix accordingly for dealing with 
		# consistency groups
		$self->{'cmd_suffix'} = "consistgrp";
	}

	my $typ = "SVCCopyServices";
	if ($OSTYPE eq "WINDOWS") {
		$self->{'cycling_mode_file'} = $ag_i18n_inc::VCSHOME . "\\bin\\" . $typ . "\\.VCS_" . $typ . "_" . $self->{'resname'} . "multicyclingmode.DO-NOT-REMOVE";
	} else {
		$self->{'cycling_mode_file'} = $ag_i18n_inc::VCSHOME . "/bin/" . $typ . "/.VCS_" .
$typ . "_" . $self->{'resname'} . "multicyclingmode.DO-NOT-REMOVE";
	}

	$self->{'entrypoint_timeout_ext'} = 86400;
	$self->{'is_timeout_extended'} = 0;
	bless($self, $class);
	return $self;
}

sub DESTROY {}

# Init other values in the global hash
sub init_values {
	my ($self) = @_;

	# Get the role of the cluster - whether master or aux
	my $ip_field;
	my $cluster_id = "";
	my $found_cluster = 0;
	my $line = "";
	my (@fields) = ();
	my $ret = 0;

	#Get SVC version using svcconfig -ver
	my ($rc, @out) = $self->_run_svcconfig_cmd($SVCVERSION);
	if (!$rc) {
		$self->{'svc_version'} = $out[0];
		VCSAG_LOGDBG_MSG(1, "SVC version is $self->{'svc_version'}", "$self->{'svc_version'}");
	}

	(@out) = ();
	if ($self->{'svc_version'} >= 5.1) {
		($rc, @out) = $self->_run_svcinfo_cmd($LSCLUSTERIP);
		$ip_field = 4;
	} else {
		($rc, @out) = $self->_run_svcinfo_cmd($LSCLUSTER);
		$ip_field = 5;
	}
	if (!$rc) {
		# Return code is 0 - So command succeeded
		foreach $line (@out) {
			(@fields) = split(':', $line);
			if ("$fields[$ip_field]" eq "$self->{'svc_cluster_ip'}") {
				# Found the cluster. Get the cluster_id.
				$cluster_id = $fields[0];
				$found_cluster = 1;
				VCSAG_LOGDBG_MSG(1, "Cluster id found using cluster IP");
				last;
			}
		}

		# If SVC management IP is different than cluster IP (eg. NAT),
		# then we get the cluster id based on the location "local".
		if ($found_cluster != 1) {
			foreach $line (@out) {
				(@fields) = split(':', $line);
				if ("$fields[2]" eq "$LOCAL") {
					$cluster_id = $fields[0];
					$found_cluster = 1;
					VCSAG_LOGDBG_MSG(1, "Cluster id found using location local");
					last;
				}	
			}
		}

	}

	(@out) = ();
	$line = "";
	(@fields) = ();
	if ($found_cluster == 1) {
		# Run the lsrc cmd on the relationship or the
		# consistency group being monitored. Match the
		# cluster_id with the ids of the master and aux clusters
		# to find out the current role of the cluster.
		($rc, @out) = $self->_run_svcinfo_cmd($LSRC);
		if (!$rc) {
		    # Command succeeded
		    foreach $line (@out) {
			@fields = split(':', $line);
			# Find out if the cluster is the master or the
			# aux cluster
			if ("$fields[1]" eq "$cluster_id") {
				if ("$fields[0]" eq "$MASTER_CLUSTER_ID") {
					$self->{'role'} = $MASTER;
					next;
				}
				elsif ("$fields[0]" eq "$AUX_CLUSTER_ID") {
					$self->{'role'} = $AUX;
					next;
				}
			}
			# Get the copy type
			elsif ("$fields[0]" eq "$COPY_TYPE") {
				# Either metro or global
				$self->{'copy_type'} = $fields[1];
			}
		    } #foreach
		} # if !rc
	} else { # if found_cluster
		VCSAG_LOG_MSG("E", "Failed to get the cluster id", 55);	
		$ret = -1;
	} # if found_cluster

	return $ret;
}

sub _run_svc_cmd {
	my ($self, @cmd) = @_;
	my $ret = 1;
	my @out = ();
	my $total_time = 0;

	VCSAG_LOGDBG_MSG(2, "Start _run_svc_cmd");

	if ($self->{'entrypoint_timeout_ext'} != 0) {
		$total_time = $self->{'entrypoint_timeout_ext'};
	} else {
		my ($oto) = split('\n', `$ag_i18n_inc::HATYPE -value SVCCopyServices OnlineTimeout`);
		$total_time = $oto;
	}

	my $cur_time = time();
	my $remaining_time = ($total_time - ($cur_time - $self->{'start_time'}));
	# Leaving 10 seconds for processing after this succeeds.
	my $count = int($remaining_time / 8) - 2;
	if ($count <= 0) {
		if ($remaining_time > 0) {
			# Try at least once since there is time remaining.
			$count = 1;
		} else {
			return $ret, @out;
		}
	}

	my $i = 1;
	my $line;
        while ($i <= $count) {
		@out = split('\n', `@cmd 2>&1`);
		$ret = $?;
		$ret = $ret >> 8;
		if (($ret != 0) && (scalar @out > 0)) {
			my $error = join(' ', @out);
			# Error given by SVC, ssh and plink are -
			# 1. "CMMVC7017E Login has failed because the maximum
			#  number of concurrent CLI sessions has been reached"
			# 2. "ssh_exchange_identification: Connection closed by
			#  remote host"
			# 3. "ssh: connect to host 10.20.134.25 port 22:
			#  Connection refused"
			# 4. "FATAL ERROR: Server unexpectedly closed network
			#  connection."
			# 5. "FATAL ERROR: Network error: Connection refused."
			if (($error =~ /^CMMVC7017E /) || ($error =~ /Connection(.*)refused/) || ($error =~ /^ssh(.*)Connection(.*)closed/) || ($error =~ /closed(.*)connection/)) {
				VCSAG_LOG_MSG("E",
					"The ssh connection failed with error - $error. Retrying ssh connection.",
					62, "$error");
			} else {
				last;
			}

			sleep (5);
			$i = $i + 1;
		} else {
			last;
		}
	}

	VCSAG_LOGDBG_MSG(2, "End _run_svc_cmd");

	return $ret, @out;
}

sub _run_svcinfo_cmd {
	my ($self, $sub_cmd, $name) = @_;
	my ($cmd) = "";
	my ($pr_cmd) = "";
	my $ret = 0;
	my @out = ();
	my $ssh_identity_file_option = "";
	my $pr_ssh_identity_file_option = "";
	my $user_name = "admin";
	my $ssh_cmd = "";

	VCSAG_LOGDBG_MSG(2, "Start _run_svcinfo_cmd");

	# Check that the ssh binary is an executable
	if (! -f $self->{'ssh_binary'} || ! -x $self->{'ssh_binary'}) {
		VCSAG_LOG_MSG("E",
			"The file $self->{'ssh_binary'} is not an executable",
			30, "$self->{'ssh_binary'}");
		$ret = -1;
		VCSAG_LOGDBG_MSG(2, "End _run_svcinfo_cmd");
		return $ret, @out;
	}

	# If it is specified, check that the id file exists
	if ((! -z $self->{'ssh_identity_file'}) && 
		($self->{'ssh_identity_file'} ne "")) {
	    if (-e $self->{'ssh_identity_file'}) {
		if (-f $self->{'ssh_identity_file'}) {
			$ssh_identity_file_option = "-i $self->{'ssh_identity_file_w_quotes'}";
			$pr_ssh_identity_file_option = "%-i $self->{'ssh_identity_file_w_quotes'}";
		}
		else {
			VCSAG_LOG_MSG("E",
				"Identity file $self->{'ssh_identity_file'} is not an ordinary file",
				31, "$self->{'ssh_identity_file'}");
		}
	    }
	    else {
		VCSAG_LOG_MSG("E",
			"Identity file $self->{'ssh_identity_file'} does not exist",
			32, "$self->{'ssh_identity_file'}");
	    }
	}

	# Check that the user name is specified. Else default it to
	# admin
	if (!-z $self->{'svc_username'}) {
		$user_name = $self->{'svc_username'};
	}

	$ssh_cmd = "$self->{'ssh_binary_w_quotes'}"." "."$ssh_identity_file_option"." -l "."$user_name"." "."$self->{'svc_cluster_ip'}";
	my $pr_ssh_cmd = "$self->{'ssh_binary_w_quotes'}"." "."$pr_ssh_identity_file_option"." %-l "."$user_name"." "."$self->{'svc_cluster_ip'}";

	if ($sub_cmd eq $LSCLUSTER) {
		# Run the lscluster cmd
		# ssh -i <id>  -l <user> 
		$cmd = $ssh_cmd." ".$SVCINFO_CMD." ".$LSCLUSTER." -delim : ";
		$pr_cmd = $pr_ssh_cmd." ".$SVCINFO_CMD." ".$LSCLUSTER." %-delim : ";
	}
	elsif ($sub_cmd eq $LSCLUSTERIP) {
		# Run the lsclusterip cmd
		# This sub-command is only available SVC 5.1 onwards
		# ssh -i <id>  -l <user>
		$cmd = $ssh_cmd." ".$SVCINFO_CMD." ".$LSCLUSTERIP." -delim : ";
		$pr_cmd = $pr_ssh_cmd." ".$SVCINFO_CMD." ".$LSCLUSTERIP." %-delim : ";
	}
	elsif ($sub_cmd eq $LSRC) {
		# Get the status of the MM/GM relationship or
		# consistency group. Need to append the appropriate cmd
		# suffix to the "lsrc" string - depends on whether we
		# are monitoring a MM/GM relationship or a consistency
		# group.
		$cmd = $ssh_cmd." ".$SVCINFO_CMD." ".$LSRC.$self->{'cmd_suffix'}." -delim : ".$self->{'name'};
		$pr_cmd = $pr_ssh_cmd." ".$SVCINFO_CMD." ".$LSRC.$self->{'cmd_suffix'}." %-delim : ".$self->{'name'};
	}
	elsif ($sub_cmd eq $LSREL) {
		# Get the details of the MM/GM relationship.
		$cmd = $ssh_cmd." ".$SVCINFO_CMD." ".$LSREL." -delim : ".$name;
		$pr_cmd = $pr_ssh_cmd." ".$SVCINFO_CMD." ".$LSREL." %-delim : ".$name;
	}
	elsif ($sub_cmd eq $LSVDISK) {
		# Run the lsvdisk cmd
		$cmd = $ssh_cmd." ".$SVCINFO_CMD." ".$LSVDISK." -delim : ".$name;
		$pr_cmd = $pr_ssh_cmd." ".$SVCINFO_CMD." ".$LSVDISK." %-delim : ".$name;
	}
	VCSAG_LOGDBG_MSG(1, "Executing cmd $pr_cmd", "$pr_cmd");
	my ($ret, @out) = $self->_run_svc_cmd($cmd);
	if ($ret != 0) {
		# If it is an SVC error, ret should be 1. In this case,
		# there should be an error message that begins with the
		# literal CMMVC.
		if ($ret == 1 && scalar @out > 0 && $out[0] !~ /^CMMVC/) {
			# Skip the msg that is not the CMMVC error msg
			shift @out;
		}
		elsif ($ret != 1) {
			# Not an SVC error. Log an appropriate message
			my $error = join(' ', @out);
			VCSAG_LOG_MSG("E",
				"Command $cmd failed for $self->{'name'} with error $error",
				24, "$cmd", "$self->{'name'}", "$error");
		}
	}

	VCSAG_LOGDBG_MSG(2, "End _run_svcinfo_cmd");
	return $ret, @out;
}

sub _run_svctask_cmd {
	my ($self, $sub_cmd, $cmd_args, $cmd_ref) = @_;
	my $cmd = "";
	my ($pr_cmd) = "";
	my $ret = 0;
	my @out = ();
	my $ssh_identity_file_option = "";
	my $pr_ssh_identity_file_option = "";
	my $user_name = "admin";
	my $ssh_cmd = "";

	VCSAG_LOGDBG_MSG(2, "Start _run_svctask_cmd");

	# Check that the ssh binary is an executable
	if (! -f $self->{'ssh_binary'} || ! -x $self->{'ssh_binary'}) {
		VCSAG_LOG_MSG("E",
			"The file $self->{'ssh_binary'} is not an executable",
			30, "$self->{'ssh_binary'}");
		$ret = -1;
		VCSAG_LOGDBG_MSG(2, "End _run_svctask_cmd");
		return $ret, @out;
	}

	# If it is specified, check that the id file exists
	if ((! -z $self->{'ssh_identity_file'}) && 
		($self->{'ssh_identity_file'} ne "")) {
	    if (-e $self->{'ssh_identity_file'}) {
		if (-f $self->{'ssh_identity_file'}) {
			$ssh_identity_file_option = "-i $self->{'ssh_identity_file_w_quotes'}";
			$pr_ssh_identity_file_option = "%-i $self->{'ssh_identity_file_w_quotes'}";
		}
		else {
			VCSAG_LOG_MSG("E",
				"Identity file $self->{'ssh_identity_file'} is not an ordinary file",
				31, "$self->{'ssh_identity_file'}");
		}
	    }
	    else {
		VCSAG_LOG_MSG("E",
			"Identity file $self->{'ssh_identity_file'} does not exist",
			32, "$self->{'ssh_identity_file'}");
	    }
	}

	# Check that the user name is specified. Else default it to
	# admin
	if (!-z $self->{'svc_username'}) {
		$user_name = $self->{'svc_username'};
	}

	#$ssh_cmd = "$self->{'ssh_binary'} $ssh_identity_file_option -l $user_name $self->{'svc_cluster_ip'}";
	$ssh_cmd = "$self->{'ssh_binary_w_quotes'}"." "."$ssh_identity_file_option"." -l "."$user_name"." "."$self->{'svc_cluster_ip'}";
	my $pr_ssh_cmd = "$self->{'ssh_binary_w_quotes'}"." "."$pr_ssh_identity_file_option"." %-l "."$user_name"." "."$self->{'svc_cluster_ip'}";

	if ($sub_cmd eq $SWITCHRC) {
		# Run the switchrc cmd
		# ssh -i <id>  -l <user> 
		$cmd = $ssh_cmd." ".$SVCTASK_CMD." ".$SWITCHRC.$self->{'cmd_suffix'}." -primary ".$self->{'role'}." ".$self->{'name'};
		$pr_cmd = $pr_ssh_cmd." ".$SVCTASK_CMD." ".$SWITCHRC.$self->{'cmd_suffix'}." %-primary ".$self->{'role'}." ".$self->{'name'};
	}
	elsif ($sub_cmd eq $STOPRC) {
		# Run the stoprc cmd
		# ssh -i <id>  -l <user> 
		$cmd = $ssh_cmd." ".$SVCTASK_CMD." ".$STOPRC.$self->{'cmd_suffix'}." ".$cmd_args." ".$self->{'name'};
		$pr_cmd = $pr_ssh_cmd." ".$SVCTASK_CMD." ".$STOPRC.$self->{'cmd_suffix'}." %".$cmd_args." ".$self->{'name'};
	}
	elsif ($sub_cmd eq $STARTRC) {
		# Run the startrc cmd
		# ssh -i <id>  -l <user> 
		$cmd = $ssh_cmd." ".$SVCTASK_CMD." ".$STARTRC.$self->{'cmd_suffix'}." ".$cmd_args." ".$self->{'name'};
		$pr_cmd = $pr_ssh_cmd." ".$SVCTASK_CMD." ".$STARTRC.$self->{'cmd_suffix'}." %".$cmd_args." ".$self->{'name'};
	}
	elsif ($sub_cmd eq $CHRC) {
		# Run the chrcrelationship or chrcconsistgrp cmd to change a
		# parameter of the relationship/consistency group.
		# ssh -i <id>  -l <user> 
		$cmd = $ssh_cmd." ".$SVCTASK_CMD." ".$CHRC.$self->{'cmd_suffix'}." ".$cmd_args." ".$self->{'name'};
		$pr_cmd = $pr_ssh_cmd." ".$SVCTASK_CMD." ".$CHRC.$self->{'cmd_suffix'}." %".$cmd_args." ".$self->{'name'};
	}

	VCSAG_LOGDBG_MSG(1, "Executing cmd $pr_cmd", "$pr_cmd");
	$$cmd_ref = $cmd;
	my ($ret, @out) = $self->_run_svc_cmd($cmd);

	VCSAG_LOGDBG_MSG(2, "End _run_svctask_cmd");
	return $ret, @out;
}

sub _run_svcconfig_cmd {
	my ($self, $sub_cmd) = @_;
	my ($cmd) = "";
	my ($pr_cmd) = "";
	my $ret = 0;
	my @out = ();
	my $ssh_identity_file_option = "";
	my $pr_ssh_identity_file_option = "";
	my $user_name = "admin";
	my $ssh_cmd = "";
	my $line;

	VCSAG_LOGDBG_MSG(2, "Start _run_svcconfig_cmd");

	# Check that the ssh binary is an executable
	if (! -f $self->{'ssh_binary'} || ! -x $self->{'ssh_binary'}) {
		VCSAG_LOG_MSG("E",
			"The file $self->{'ssh_binary'} is not an executable",
			30, "$self->{'ssh_binary'}");
		$ret = -1;
		VCSAG_LOGDBG_MSG(2, "End _run_svcconfig_cmd");
		return $ret, @out;
	}

	# If it is specified, check that the id file exists
	if ((! -z $self->{'ssh_identity_file'}) && 
		($self->{'ssh_identity_file'} ne "")) {
	    if (-e $self->{'ssh_identity_file'}) {
		if (-f $self->{'ssh_identity_file'}) {
			$ssh_identity_file_option = "-i $self->{'ssh_identity_file_w_quotes'}";
			$pr_ssh_identity_file_option = "%-i $self->{'ssh_identity_file_w_quotes'}";
		}
		else {
			VCSAG_LOG_MSG("E",
				"Identity file $self->{'ssh_identity_file'} is not an ordinary file",
				31, "$self->{'ssh_identity_file'}");
		}
	    }
	    else {
		VCSAG_LOG_MSG("E",
			"Identity file $self->{'ssh_identity_file'} does not exist",
			32, "$self->{'ssh_identity_file'}");
	    }
	}

	# Check that the user name is specified. Else default it to
	# admin
	if (!-z $self->{'svc_username'}) {
		$user_name = $self->{'svc_username'};
	}

	$ssh_cmd = "$self->{'ssh_binary_w_quotes'}"." "."$ssh_identity_file_option"." -l "."$user_name"." "."$self->{'svc_cluster_ip'}";
	my $pr_ssh_cmd = "$self->{'ssh_binary_w_quotes'}"." "."$pr_ssh_identity_file_option"." %-l "."$user_name"." "."$self->{'svc_cluster_ip'}";

	if ($sub_cmd eq $SVCVERSION) {
		# Run the svcconfig cmd
		# ssh -i <id>  -l <user> 
		$cmd = $ssh_cmd." ".$SVCCONFIG_CMD." ".$SVCVERSION;
		$pr_cmd = $pr_ssh_cmd." ".$SVCCONFIG_CMD." %".$SVCVERSION;
	}

	VCSAG_LOGDBG_MSG(1, "Executing cmd $pr_cmd", "$pr_cmd");
	# The output of svcconfig -ver sometimes also includes the command.
	# So skipping until last line, which has the output i.e. SVC version.
	my ($ret, @out1) = $self->_run_svc_cmd($cmd);
	foreach $line (@out1) {
		chomp($line);
		@out = $line;
	}

	VCSAG_LOGDBG_MSG(2, "End _run_svcconfig_cmd");
	return $ret, @out;
}

#
# Get the current state of replication for the specified relationship or
# consistency group. The state can be one of the valid states defined
# above or invalid, if we get back something that is not defined in the
# list of valid states defined in this module.
#
sub get_state {
	my ($self) = @_;
	my $state = "";
	my ($rc, @out) = $self->_run_svcinfo_cmd($LSRC);
	if (!$rc) {
	    # Command succeeded
	    foreach my $line (@out) {
		if ($line =~ /^state/) {
			my (@stateinfo) = split(':', $line);
			$state = $stateinfo[1];
			last;
		}
	    }
	}

	my $state_valid = 0;
	foreach my $stateval (@VALID_STATES) {
		if ($stateval eq $state) {
			$state_valid = 1;
			last;
		}
	}

	if (!$state_valid) {
		$state = "Invalid";
	}

	return $state;
}

#
# Get the cycling mode of replication for the specified relationship or
# consistency group. The cycling mode can be "none" or "multi".
#
sub get_cycling_mode {
	my ($self) = @_;
	my $cycling_mode = "";
	my ($rc, @out) = $self->_run_svcinfo_cmd($LSRC);
	if (!$rc) {
	    # Command succeeded
	    foreach my $line (@out) {
		if ($line =~ /^cycling_mode/) {
			my (@cycling_info) = split(':', $line);
			$cycling_mode = $cycling_info[1];
			last;
		}
	    }
	}

	return $cycling_mode;
}

#
# Return the status of sync. If the "sync" attribute for the replication
# relationship or consistency group indicates that the master and aux
# are in sync wrt data, then this routine will return a 1 (in-sync),
# else a 0. The value of the "sync" attribute can be "" or "in_sync" to
# indicate that the data is in sync between the master and aux clusters.
#
sub get_sync_status {
	my ($self) = @_;
	my $sync = "";

	my ($rc, @out) = $self->_run_svcinfo_cmd($LSRC);
	if (!$rc) {
	    # Command succeeded
	    foreach my $line (@out) {
		if ($line =~ /^sync/) {
			my (@syncinfo) = split(':', $line);
			$sync = $syncinfo[1];
			last;
		}
	    }
	}

	my $in_sync = 0;
	foreach my $insync_val (@INSYNC_STATUS) {
		if ($insync_val eq $sync) {
			$in_sync = 1;
			last;
		}
	}

	return $in_sync, $sync;
}

#
# Return the value of the "status" field in the output of svcinfo lsrc
# cmd.
#
sub get_link_status {
	my ($self) = @_;
	my $status = "";

	my ($rc, @out) = $self->_run_svcinfo_cmd($LSRC);
	if (!$rc) {
	    # Command succeeded
	    foreach my $line (@out) {
		if ($line =~ /^status/) {
			my (@statusinfo) = split(':', $line);
			$status = $statusinfo[1];
			last;
		}
	    }
	}

	return $status;
}

#
# Checks if the local cluster is the Primary in the replication. If it
# is, then the routine returns 1 else a 0.
#
sub is_primary {
	my ($self) = @_;
	my $primary = "";
	my $is_primary = 0;
	my ($rc, @out) = $self->_run_svcinfo_cmd($LSRC);
	if (!$rc) {
	    # Command succeeded
	    foreach my $line (@out) {
		if ($line =~ /^primary/) {
			my (@primaryinfo) = split(':', $line);
			$primary = $primaryinfo[1];
			last;
		}
	    }
	}
	if ($primary eq $self->{'role'}) {
		$is_primary = 1;
	}
	return $is_primary;
}

sub get_action_required {
	my ($self) = @_;
	my $is_primary = 0;
	my $action = 0;
	my $state = "";
	my $ret = 0;

	$state = $self->get_state();

	if (($state eq "Invalid") || ($state eq "empty")) {
		VCSAG_LOG_MSG("E",
			"State of replication ($state) for $self->{'name'} is invalid",
			17, "$state", "$self->{'name'}");
		$action = $ACTION_DONOTHING_ISOFFLINE;
		return $action;
	}

	$is_primary = $self->is_primary();
	if ($is_primary) {
		# Nothing to be done - the current cluster is already
		# the Primary cluster. Do we want to start the
		# replication if it is stopped here?
		VCSAG_LOG_MSG("N",
			"SVC cluster $self->{'svc_cluster_ip'} is the Primary for $self->{'name'}. So read/write access is already enabled to the cluster.",
			18, "$self->{'svc_cluster_ip'}", "$self->{'name'}");
		$action = $ACTION_DONOTHING_ISONLINE;
		return $action;
	}

	if ($state eq $STATE_INCONSISTENT_COPYING) {
		# Wait for the copy to complete
		$ret = $self->wait_for_copy_complete($state);
		if ($ret == 0) {
			# State is no longer InconsistentCopying.
			# Get the current state and proceed
			$state = $self->get_state();
		}
		elsif ($ret == 1) {
			# Timedout waiting for the relationship or
			# consistency group to move out of the
			# inconsistent copying state
			VCSAG_LOG_MSG("E",
				"Timed out waiting for background copy to complete on SVC cluster $self->{'svc_cluster_ip'} for $self->{'name'}",
				19, "$self->{'svc_cluster_ip'}",
				"$self->{'name'}");
			$action = $ACTION_DONOTHING_ISOFFLINE;
			return $action;
		}
	}

	if (($state eq $STATE_IDLING) ||
		($state eq $STATE_IDLING_DISCONNECTED)) {
		# Both master and aux will have read/write access. So nothing
		# needs to be done here
		VCSAG_LOG_MSG("N", 
			"read/write access is already enabled on SVC cluster $self->{'svc_cluster_ip'} for $self->{'name'} because the state of the SVC cluster is $state",
			20, "$self->{'svc_cluster_ip'}",
			"$self->{'name'}", "$state");
		$action = $ACTION_DONOTHING_ISONLINE;
	}
	elsif ($state eq $STATE_CONSISTENT_SYNCHRONIZED) {
		# We need to do a switch to enable rw access to the SVC
		# cluster as well as become the Primary in the
		# replication relationship
		$action = $ACTION_SWITCHRC;
	}
	elsif ($state eq $STATE_CONSISTENT_STOPPED) {
		# Dont do anything, check if StopTakeover is 1. If yes, 
		# then do a switchrc if data is in sync, else do a 
		# stoprc -access.
		$action = $ACTION_STOPTAKEOVER;
	}
	elsif ($state eq $STATE_CONSISTENT_DISCONNECTED) {
		$action = $ACTION_DISCTAKEOVER;
	}
	elsif ($state eq $STATE_CONSISTENT_COPYING) {
		# Dont do anything, check if CopyTakeover is 1. If yes, 
		# then do a switchrc by changing cycling mode, else do a 
		# stoprc -access.
		$action = $ACTION_COPYTAKEOVER;
	}
	elsif (($state eq $STATE_INCONSISTENT_DISCONNECTED) ||
		($state eq $STATE_INCONSISTENT_STOPPED)) {
		# Dont do anything here. Even if AutoTakeover = 1, we
		# cannot takeover cos the data on this cluster is in an
		# inconsistent state
		VCSAG_LOG_MSG("E", 
			"Cannot enable read/write access on SVC cluster $self->{'svc_cluster_ip'} for $self->{'name'} because the data is in an inconsistent state and the state of the SVC cluster $state",
			21, "$self->{'svc_cluster_ip'}",
			"$self->{'name'}", "$state");
		$action = $ACTION_DONOTHING_ISOFFLINE;

	}

	return $action;
}

#
# This subroutine waits for a period of time for the state of the
# replication to move out of INCONSISTENT_COPYING. The period of time
# that this routine waits is the remaining time left for online ep to
# timeout minus a grace period of one second. The subroutine sleeps 5
# seconds, checks the status of replication. If it is still
# INCONSISTENT_COPYING, then stay in the loop. Else get out of the loop
# and return from the function.
# Returns 0 on success and 1 on failure
#
sub wait_for_copy_complete {
	my ($self, $prev_state) = @_;
	my ($oto) = split('\n', `$ag_i18n_inc::HATYPE -value SVCCopyServices OnlineTimeout`);
	my $cur_time = time();
	my $remaining_time = ($oto - ($cur_time - $self->{'start_time'}));
	my $count = int($remaining_time / 5) - 1;
	if ($count <= 0) {
		return 1;
	}

	my $i = 1;
	my $state = "";
	while ($i <= $count) {
		$state = $self->get_state();
		if ($state ne $prev_state) {
			VCSAG_LOG_MSG("N",
				"State of replication for $self->{'name'} changed from $prev_state to $state",
				22, "$self->{'name'}", "$prev_state",
				"$state");
			last;
		}
		# sleep 5 seconds
		sleep (5);
		$i = $i + 1;
	}

	return 0;
}

#
# THis subroutine waits for a period of time for the state of the
# replication to become IDLING or IDLING_DISCONNECTED. The period of time
# that this routine waits is the remaining time left for online ep to
# timeout minus a grace period of one second. The subroutine sleeps 5
# seconds, checks the status of replication. If it is still
# not IDLING or IDLING_DISCONNECTED, then stay in the loop. Else get 
# out of the loop and return from the function.
# Returns 0 on success and 1 on failure
#
sub wait_for_idling {
	my ($self, $prev_state) = @_;
	my ($oto) = split('\n', `$ag_i18n_inc::HATYPE -value SVCCopyServices OnlineTimeout`);
	my $cur_time = time();
	my $remaining_time = ($oto - ($cur_time - $self->{'start_time'}));
	my $count = int($remaining_time / 5) - 1;
	if ($count <= 0) {
		return 1;
	}

	my $i = 1; 
	my $state = "";
	while ($i <= $count) {
		$state = $self->get_state();
		if (($state eq $STATE_IDLING) ||
			($state eq $STATE_IDLING_DISCONNECTED)) {
			VCSAG_LOG_MSG("N",
				"State of replication for $self->{'name'} changed from $prev_state to $state",
				22, "$self->{'name'}", "$prev_state", "$state");
			last;
		}
		# sleep 5 seconds
		sleep (5);
		$i = $i + 1;
	}

	return 0;
}

#
# A subroutine that invokes the "switchrc" command to switch the Primary
# between the master and aux clusters. The function will return a 0 on
# success and 1 on failure and 2 if the current state of the replication
# does not allow a switchrc command - so we can try to do a stop
# takeover.
# This subroutine should be called only if the state of the replication
# is ConsistentSync or ConsistentStopped with Primary and Secondary in
# sync. In all other states and conditions, the switchrc command will
# fail.
#
sub switchrc {
	my ($self) = @_;

	my $ret = 0;
	my $cmd = "";

	my ($rc, @out) = $self->_run_svctask_cmd($SWITCHRC, "", \$cmd);
	if ($rc != 0) {
	    # Error running the svctask command
	    VCSAG_LOG_MSG("E",
		"Command $cmd failed for $self->{'name'} with error $out[0]",
		24, "$cmd", "$self->{'name'}",
		"$out[0]");

	    $ret = 1;
	    
	    if ($out[0] =~ /^CMMVC/) {
		my @error_fields = split(' ', $out[0]);
		# Get the error reason
		if (($error_fields[0] eq "CMMVC5973E") ||
			($error_fields[0] eq "CMMVC5978E")) {
			# switchrc failed because Primary and
			# Secondary are not in sync - attempt a
			# stop takeover, if possible.
			$ret = 2;
		}
	    }
	}
	else {
		# Check that the local SVC cluster is the primary
		my $is_primary = $self->is_primary();
		if ($is_primary) {
		# success
			VCSAG_LOG_MSG("N",
			"Command $cmd succeeded for $self->{'name'}",
			25, "$cmd", "$self->{'name'}");
		}
		else {
			$ret = 1;
			VCSAG_LOG_MSG("E",
			"Command $cmd failed for $self->{'name'}",
			26, "$cmd", "$self->{'name'}");
		}
	}
	return $ret;
}

sub takeover {
	my ($self, $takeover_type) = @_;
	my $ret = 0;
	my $rc = 0;
	my $cmd = "";
	my $cmd_opt = "";
	my (@out) = ();

	my $state = $self->get_state();
	if ((($takeover_type == $DISCONNECT_TAKEOVER) &&  
		($state ne $STATE_CONSISTENT_DISCONNECTED)) && 
	    (($takeover_type == $STOP_TAKEOVER) &&
		($state ne $STATE_CONSISTENT_STOPPED)) &&
	    (($takeover_type == $COPY_TAKEOVER) && 
		($state ne $STATE_CONSISTENT_COPYING))) {

		# Takeover type and state do not match
		# Cannot do a takeover
		$ret = 1;
		return $ret;
	}

	# Come here only if:
	# DisconnectTakeover and state = consistent_disconnected
	# OR
	# StopTakeover and state = consistent_stopped

	if (($takeover_type == $DISCONNECT_TAKEOVER) &&
		($self->{'copy_type'} eq $COPYTYPE_METROMIRROR)) {
		# ONLY in the the case of a disconnected state for a
		# metro mirror relationship or consistency group:
		# Make sure that the Primary and Secondary are in sync
		# before making read/write access available on the Secondary
		my ($in_sync, $sync_status) = $self->get_sync_status();
		if (!$in_sync) {
			# Data is not in sync and copy type is metro
			# mirror - Dont do a takeover
			VCSAG_LOG_MSG("E",
				"Cannot enable read/write access for $self->{'name'} with the role $self->{'role'} because copy type is $self->{'copy_type'} and the sync status is $sync_status",
				23, "$self->{'name'}", "$self->{'role'}",
				"$self->{'copy_type'}", "$sync_status");
			$ret = 1;
			return $ret;
		}
	}

	if ($takeover_type == $STOP_TAKEOVER) {
		# Replication is in Consistent_stopped state.
		# Check if the data is "in_sync"
		my ($in_sync, $sync_status) = $self->get_sync_status();
		VCSAG_LOGDBG_MSG(1, "Sync status is $sync_status for $self->{'name'}",
			"$sync_status", "$self->{'name'}");
		if ($in_sync) {
			# Should be possible to do a switchrc
			$ret = $self->switchrc();
			if ($ret == 1) {
				# switch failed due to unknown reasons
				return $ret;
			}
			elsif ($ret == 2) {
				# Data is not in sync.
				# So fall through and try the stoprc -access
				# cmd to see if we can salvage anything here.
			}
			elsif ($ret == 0) {
				# switchrc succeeded
				return $ret;
			}
		}
	}

	# The action COPY_TAKEOVER and the state consistent_copying, is
	# applicable only when cycling mode is multi. If we do a regular
	# switch, the data could be lost. However, with multi cycling mode,
	# the data is always consistent (except before 1st copy is completed).
	# The sync attribute always remains out_of_sync. A temporary
	# workaround suggested by IBM or SVC 6.3 is to change the cycling mode
	# to none, and then perform the regular switch. 

	if ($takeover_type == $COPY_TAKEOVER) {
		VCSAG_LOGDBG_MSG(1, "Attempting command $STOPRC for $self->{'name'}; current state $state", "$STOPRC", "$self->{'name'}", "$state");	
		
		($rc, @out) = $self->_run_svctask_cmd($STOPRC, $cmd_opt, \$cmd);
		
		if ($rc != 0) {
			# Error condition
			my $error = join(' ', @out);
			VCSAG_LOG_MSG("E",
				"Command $cmd failed for $self->{'name'} with error $error",
				24, "$cmd", "$self->{'name'}", "$error");
			$ret = 1;
			return $ret;
		}

		$state = $self->get_state();
		if ($state ne $STATE_CONSISTENT_STOPPED) {
			VCSAG_LOG_MSG("E",
				"The state expected was $STATE_CONSISTENT_STOPPED but it is $state. Manual intervention may be required",
				56, "$STATE_CONSISTENT_STOPPED", "$state");
				$ret = 1;
				return $ret;
		}

		my $cycling_mode = $self->get_cycling_mode();

		if ($cycling_mode eq "multi") {
			VCSAG_LOGDBG_MSG(1, "Saving cycling mode as 'multi' in file");
			open(lfh, ">$self->{'cycling_mode_file'}");
			close(lfh);
		} else {
			# The support for consistent_copying is only added for
			# cycling mode "multi".
			VCSAG_LOG_MSG("E",
				"The state $state is not expected with cycling mode $cycling_mode", 61, "$state", "$cycling_mode");
		}

		# The stoprc command worked.
		# Now change the cycling mode from multi to none.
		$cmd_opt = "-cyclingmode none";
		VCSAG_LOGDBG_MSG(1, "Attempting command $CHRC for $self->{'name'}; current state $state", "$CHRC", "$self->{'name'}", "$state");

		($rc, @out) = $self->_run_svctask_cmd($CHRC, $cmd_opt, \$cmd);
		
		if ($rc != 0) {
			# Error condition
			my $error = join(' ', @out);
			VCSAG_LOG_MSG("E",
				"Command $cmd failed for $self->{'name'} with error $error",
				24, "$cmd", "$self->{'name'}", "$error");
			$ret = 1;
			return $ret;
		}

		# startrc with force.
		$cmd_opt = "-force";
		VCSAG_LOGDBG_MSG(1, "Attempting command $STARTRC with force for $self->{'name'}; current state $state", "$STARTRC", "$self->{'name'}", "$state");

		($rc, @out) = $self->_run_svctask_cmd($STARTRC, $cmd_opt, \$cmd);
		if ($rc != 0) {
			# Error condition
			my $error = join(' ', @out);
			VCSAG_LOG_MSG("E",
				"Command $cmd failed for $self->{'name'} with error $error",
				24, "$cmd", "$self->{'name'}", "$error");

				# So fall through and try the stoprc -access
				# cmd to see if we can salvage anything here.
		} else {
			# The startrc with force command worked.
			# Wait for state to change to consistent synchronized
			$state = $self->get_state();
			if ($state ne $STATE_CONSISTENT_SYNCHRONIZED) {
				$ret = $self->wait_for_copy_complete($state);
				if ($ret == 0) {
					# Get the current state and proceed
					$state = $self->get_state();
				}
				elsif ($ret == 1) {
					# Timedout waiting for the relationship
					# or consistency group to move out of
					# the copying state.
					VCSAG_LOG_MSG("E",
						"Attempted background copy to synchronize when cycling mode is set to 'multi'. Timed out waiting for background copy to complete on SVC cluster $self->{'svc_cluster_ip'} for $self->{'name'}",
						57, "$self->{'svc_cluster_ip'}",
						"$self->{'name'}");
					# So fall through and try the stoprc
					# -access cmd to see if we can salvage
					# anything here.
				}
			}

			# switchrc now.
			# Check if the data is "in_sync"
			my ($in_sync, $sync_status) = $self->get_sync_status();
			VCSAG_LOGDBG_MSG(1, "Sync status is $sync_status for $self->{'name'}",
				"$sync_status", "$self->{'name'}");
			if ($in_sync) {
				# Should be possible to do a switchrc
				$ret = $self->switchrc();
			}

			# The switchrc command worked. Else we will fall
			# through to attempt stoprc -access.
			if ($ret == 0) {

				# stoprc without force now.
				$cmd_opt = "";
				VCSAG_LOGDBG_MSG(1, "Attempting command $STOPRC for $self->{'name'}; current state $state", "$STOPRC", "$self->{'name'}", "$state");	
				
				($rc, @out) = $self->_run_svctask_cmd($STOPRC, $cmd_opt, \$cmd);
				
				if ($rc != 0) {
					# Error condition
					my $error = join(' ', @out);
					VCSAG_LOG_MSG("E",
						"Command $cmd failed for $self->{'name'} with error $error",
						24, "$cmd", "$self->{'name'}", "$error");
				}

				# The stoprc command worked.

				# Now change the cycling mode from none to multi.
				$cmd_opt = "-cyclingmode multi";
				VCSAG_LOGDBG_MSG(1, "Attempting command $CHRC for $self->{'name'}; current state $state", "$CHRC", "$self->{'name'}", "$state");

				($rc, @out) = $self->_run_svctask_cmd($CHRC, $cmd_opt, \$cmd);
				
				if ($rc != 0) {
					# Error condition
					my $error = join(' ', @out);
					VCSAG_LOG_MSG("E",
						"Command $cmd failed for $self->{'name'} with error $error",
						24, "$cmd", "$self->{'name'}", "$error");
				}

				# Now that cycling mode is set to "multi", we
				# will delete the multi file created to keep
				# this state across crash/restart.
				if (-f "$self->{'cycling_mode_file'}") {
					unlink("$self->{'cycling_mode_file'}");
				}

				# startrc now.
				$cmd_opt = "";
				VCSAG_LOGDBG_MSG(1, "Attempting command $STARTRC for $self->{'name'}; current state $state", "$STARTRC", "$self->{'name'}", "$state");

				($rc, @out) = $self->_run_svctask_cmd($STARTRC, $cmd_opt, \$cmd);
				if ($rc != 0) {
					# Error condition
					my $error = join(' ', @out);
					VCSAG_LOG_MSG("E",
						"Command $cmd failed for $self->{'name'} with error $error",
						24, "$cmd", "$self->{'name'}", "$error");

						# So fall through and try the stoprc -access
						# cmd to see if we can salvage anything here.
				}

				# The startrc command worked.
				return $ret;
			}
		}
	}

	VCSAG_LOGDBG_MSG(1, "Attempting command $STOPRC with $ACCESS_OPT for $self->{'name'}; current state $state",
		"$STOPRC", "$ACCESS_OPT", "$self->{'name'}", "$state");
	# Do a stoprc with the -access option

	($rc, @out) = $self->_run_svctask_cmd($STOPRC, $ACCESS_OPT, \$cmd);

	if ($rc != 0) {
		# Error condition
		my $error = join(' ', @out);
		VCSAG_LOG_MSG("E",
			"Command $cmd failed for $self->{'name'} with error $error",
			24, "$cmd", "$self->{'name'}", "$error");
		$ret = 1;
		return $ret;
	}

	# The stoprc -access command worked.
	# Wait for state to go to Idling or Idling
	# Disconnected
	VCSAG_LOGDBG_MSG(1, "Waiting for state of $self->{'name'} to go to Idling or Idling Disconnected after running command $cmd",
		"$self->{'name'}", "$cmd");

	$ret = $self->wait_for_idling($state);
	if ($ret) {
		# Timedout waiting for the state of
		# replication to go to idling or idling
		# disconnected
		VCSAG_LOG_MSG("E",
			"Timed out while waiting for state of replication for $self->{'name'} to go to Idling or Idling Disconnected after running the command $cmd",
			28, "$self->{'name'}", "$cmd");
	}
	else {
		# $ret == 0 i.e. success
		VCSAG_LOG_MSG("N",
			"$cmd successfully enabled read/write access for $self->{'name'} on the $self->{'role'} SVC cluster",
				13, "$cmd", "$self->{'name'}",
				"$self->{'role'}");
	}

	return $ret;
}

sub update {
	my ($self, $primary) = @_;
	my $cmd_opt = "";
	my $ret = 0;
	my $cmd = "";
	my $rc = 0;

	if ($primary ne "") {
		$cmd_opt = "-primary $primary -force";
	}
	else {
		$cmd_opt = "-force";
	}

	my ($rc, @out) = $self->_run_svctask_cmd($STARTRC, $cmd_opt, \$cmd);
	if ($rc != 0) {
		# Error condition
		my $error = join(' ', @out);
		VCSAG_LOG_MSG("E",
			"Command $cmd failed for $self->{'name'} with error $error",
			24, "$cmd", "$self->{'name'}", "$error");
	}

	return $rc, @out;
}

#
# Used to check the configuration of the SSH attributes. In particular,
# the subroutine ensures that 
# 	1. The SSHBinary attribute specifies an existing executable file
#	2. If specified, the SSHPathToIDFile specifies an existing
#	ordinary file.
# Returns 0 if attributes seem correct
# Otherwise, it returns 1.
#
sub verify_attributes {

	my ($self) = @_;
	my $ret = 0;

	# Check that the ssh binary is an executable
	if (! -f $self->{'ssh_binary'} || ! -x $self->{'ssh_binary'}) {
		VCSAG_LOG_MSG("E",
			"The file $self->{'ssh_binary'} is not an executable",
			30, "$self->{'ssh_binary'}");
		$ret = 1;
		return $ret;
	}

	# If it is specified, check that the id file exists
	if (! -z $self->{'ssh_identity_file'}) {
	    if (-e $self->{'ssh_identity_file'}) {
		if (!-f $self->{'ssh_identity_file'}) {
			VCSAG_LOG_MSG("E",
				"Identity file $self->{'ssh_identity_file'} is not an ordinary file",
				31, "$self->{'ssh_identity_file'}");
			$ret = 1;
		}
	    }
	    else {
		VCSAG_LOG_MSG("E",
			"Identity file $self->{'ssh_identity_file'} does not exist",
			32, "$self->{'ssh_identity_file'}");
		$ret = 1;
	    }
	}

	return $ret;
}

#
# The following subroutine returns the status information that will be
# displayed in the ResourceInfo attribute. Currently, this subroutine
# returns the fields: primary, state, freeze time, status and sync from
# the lsrc command output for the replication relationship or
# consistency group.
# Returns 0 on successfully capturing the values of the said fields
# above; 1 otherwise
#
sub get_replication_status {
	my ($self) = @_;
	my $ret = 0;
	my $status = "";
	my @out = ();

	($ret, @out) = $self->_run_svcinfo_cmd($LSRC);
	if (!$ret) {
	    # Command succeeded
	    foreach my $line (@out) {
		# remove \n at the end of the line
		chomp($line);
		$status = $status.$line." ";
	    }
	    if ($status ne "") {
		    # Remove the trailing blank - not really needed
		    chop($status);
	    }
	}
	return $ret, $status;
}

sub extend_entrypoint_timeout {
	my ($self) = @_;
	my $ret = 1;
	
	my $fileName = "$self->{'resname'}" . ".tmo";
	my $relative_tmp_path = $ag_i18n_inc::VCSLOG . "/log/tmp/";
	

	if ( $^O =~ /^MSWin/i) {
		$relative_tmp_path = $ag_i18n_inc::VCSHOME . "\\log\\tmp\\";
	}

	my $tmo_extension_file =  $relative_tmp_path . $fileName;

	open TMOEXTENSIONFILE , ">$tmo_extension_file" || die ("Cannot open $tmo_extension_file file!\n");
	print TMOEXTENSIONFILE "$self->{'entrypoint_timeout_ext'}";
	close TMOEXTENSIONFILE;

	# We need to keep track of the timeout to retry ssh.
	$self->{'is_timeout_extended'} = 1;
	$ret = 0;
	return $ret;
}

#
# Need to run the subroutine only on the CVM master when the agent is
# running in a RAC environment
#
sub update_disks_udid {
	my ($self) = @_;

	my $ret = 0;
	my $rc = 0;
	my ($cnt) = 0;
	my (@rel_name);
	my (@disks);
	my (@out) = ();
	my (@fields) = ();
	my (@vdisk) = ();
	my (@vdisk_uid) = ();
	my ($vd);
	my (%vxdisk) = ();
	my (@cmds) = ();
	my $line = "";
	my $searchdev = $self->{'role'}."_vdisk_name";

	if ($self->{'rac'} == 1) {
		# We dont have Firedrill support in a RAC env for SVC
		# yet - so dont do anything
		return 0;
	}

	my $sgname = $self->{'sgname'};
	my (@listdgs) = split('\n',
		`$ag_i18n_inc::HARES -list Type=DiskGroup Group=$sgname -localclus`);
	my $listdg_ret = $?;

	if ($listdg_ret != 0) {
		# This routine applies *only* for VxVM diskgroups. Since there
		# are no DiskGroup resources in the group that the SVC resource
		# belongs to, we can assume that VxVM is not the volume manager
		# used to manage the disks.
		VCSAG_LOGDBG_MSG(2,
			"Not updating disks udid since VxVM is not the underlying volume manager");
		return 0;
	}

	# If consistgrp, the find the relationships.	
	if ($self->{'is_consistency_group'} == 1) {

		# svcinfo lsrcconsistgrp -delim : <consistency group name>
		$line = "";
		($rc, @out) = $self->_run_svcinfo_cmd($LSRC);	
		if (!$rc) {
			# Command succeeded
			foreach $line (@out) {
				# Find relationships
				if ($line =~ /^RC_rel_name/) {
					@fields = split(':', $line);
					$rel_name[$cnt] = $fields[1];
					$cnt++;
					next;
				}
			}
		}
	} else {
		$rel_name[0] = $self->{'name'};
	}	
	
	@out = ();
	$cnt = 0;
	my $rel;
	foreach $rel (@rel_name) {
		#
		# Get vdisks from relationship or rcconsistgrp
		# svcinfo lsrcrelationship -delim : <relationship name>
		#
		$line = "";
		($rc, @out) = $self->_run_svcinfo_cmd($LSREL, $rel);	
		
		if (!$rc) {
			# Command succeeded
			foreach $line (@out) {
				# Find vdisk
				if ($line =~ /^${searchdev}/) {
					@fields = split(':', $line);
					$vdisk[$cnt] = $fields[1];
					$cnt++;
					last;;
				}
			}
		}
		next;
	}
		
	# Get uid of vdisk
	my $vdisk_cmd = "";
	$cnt = 0;
	foreach $vd (@vdisk) {
		# Get uid of vdisks
		$line = "";
		@out = ();
		# svcinfo lsvdisk -delim : <virtual disk name>
		($rc, @out) = $self->_run_svcinfo_cmd($LSVDISK, $vd);	
		
		if (!$rc) {
			# Command succeeded
			foreach $line (@out) {
				# Find vdisk uid
				if ($line =~ /^vdisk_UID/) {
					@fields = split(':', $line);
					$vdisk_uid[$cnt] = $fields[1];
					$cnt++;
					last;
				}
			}
		}
		next;
	}


	# Find vxvm disks which match the uids of the vdisks. The VxVM disk's
	# lun serial number field has the uid of the vdisk.
	# First get all the disks and their lun serial number
	my %vmdisk;
	my $uid = "";
	my $disk = "";
	my @disk_info;
	my @lun_info;
	my $len = 0;

	@out = ();
	$line = "";
	($rc, @out) = $self->_run_cmd("$VXDISK -p list");
	if (!$rc) {
		# Command succeeded
		foreach $line (@out) {
			if ($line =~ /^DISK\s+/) {
				@disk_info = split(':', $line);
				$disk = ltrim($disk_info[1]);
				chomp($disk);
			} elsif ($line =~ /^LUN_SERIAL_NO\s+/) {
				@lun_info = split(':', $line);
				$uid = ltrim($lun_info[1]);
				chomp($uid);
				# When SVC disk is exported to client LPAR as
				# virtual SCSI device (through VIOS), the LUN
				# SERIAL Number is not exactly the vdisk UID. It
				# has a prefix and suffix. We extract the 16
				# bytes vdisk UID.
				$len = length($uid);
				if ($len > 32) {
					$uid = substr($uid, 5, 32);
					$len = length($uid);
				}
				$vmdisk{"$uid"} = "$disk";
				$disk = "";
				$uid = "";
			}
		}
	}

	# Now get the disks with vdisk's uid matching in the disk array
	$cnt = 0;
	foreach $vd (@vdisk_uid) {
		#
		# Get the VxVM disk which has last 32 characters 
		# of the disk udid matching uid
		#
		$disks[$cnt] = $vmdisk{"$vd"};
		$cnt++;
	}

	# Update udid of the vxvm disks.
	my (@lines) = split('\n', `$VXDISK -eq list`);
	foreach my $l (@lines) {
		my (@fields) = split('\s+', $l);
		$vxdisk{$fields[$#fields]} = $fields[0];
	}

	push(@cmds, "$VXDCTL enable");
	$ret = ar_run_cmds(@cmds);

	foreach my $disk (@disks) {
		(@cmds) = ();
		if ($vxdisk{$disk}) {
			$disk = $vxdisk{$disk};
		}
		push(@cmds, "$VXDISK clearimport $disk");
		push(@cmds, "$VXDISK updateudid $disk");
		$ret = ar_run_cmds(@cmds);
		if ($ret) {
			VCSAG_LOG_MSG("E",
			"Could not update udid of the disk $disk; exiting;", 47, "$disk");
		}
	}

	return $ret;
}

sub ar_do {
	my ($do, $line, $ret);
	my $exit_code = 0;
	($do) = (@_);
	VCSAG_LOG_MSG("N", "Issuing command: $do", 50, "$do");
	open(INPUT,"$do 2>&1 |") || die("Failed to invoke $do");
	while ($line = <INPUT>) { 
		chomp($line); 
		$ret = ($ret) ? "$ret\n$line" : $line; 
	}
	close INPUT;
	$exit_code = $?;
	if ($exit_code) {
		my $status = $exit_code / 256;
		VCSAG_LOG_MSG("E", "Command $do failed, return code: $status, output: \n$ret\n", 48, "$do", "$status", "$ret");
	} else {
		VCSAG_LOG_MSG("N", "Command $do succeeded", 49, "$do");
	}
	return $exit_code;
}

sub ar_run_cmds {
	my (@cmds) = @_;
	my $ret = 0;
	foreach my $cmd (@cmds) {
		$ret = ar_do($cmd);
		if ($ret) { last; }
	}
	@cmds = ();
	return $ret;
}

sub _run_cmd {
	my ($self, $cmd) = @_;
	my $ret = 0;
	my (@out) = ();

	VCSAG_LOGDBG_MSG(1, "Executing cmd $cmd");

	@out = `$cmd`;
	$ret = $?;

	VCSAG_LOGDBG_MSG(1, "End _run_cmd");

	return $ret, @out;
}

# Left trim function to remove leading whitespace
sub ltrim($)
{
	my $string = shift;
	$string =~ s/^\s+//;
	return $string;
}

#
# The subroutine is called ONLY in a RAC environment (i.e. if the group
# for which a FireDrill is being run is a parallel group).
# This routine checks if the node is the CVM MASTER
# It returns:
# 	-1 on error
# 	1 if the current node is the MASTER
# 	0 if the current node is not the MASTER
sub is_cvm_master {
	my (@lines) = split('\n', `$VXDCTL -c mode`);
	if ($#lines == -1) {
		# i.e. no output generated for the cmd
		VCSAG_LOG_MSG("E",
"Command [$VXDCTL -c mode] failed; Could not determine vxconfigd mode; exiting;",
			52, "$VXDCTL");
		return -1;
	}

	# The first line contains info abt the cluster state and
	# the role of the node in the cluster: MASTER/SLAVE
	my $l = "";
	my $line = "";
	foreach $l (@lines) {
		if ($l =~ /^mode:/) {
			$line = $l;
			last;
		}
		next;
	}

	# Make sure we have the output line that contains the
	# mode information
	my (@fields) = split('\s+', $line);
	if ($fields[0] ne "mode:") {
		# we have the wrong line
		VCSAG_LOG_MSG("E",
"Could not determine vxconfigd mode from the output of the command [$VXDCTL -c mode]; exiting;",
			53, "$VXDCTL");
		return -1;
	}

	# From the manpage for vxdctl, clustered_state is one of:
	# 	cluster inactive,
	# 	cluster active  -  role not set,
	# 	cluster active - MASTER,
	# 	cluster active - SLAVE
	if ($line =~ /cluster inactive/ || $line =~ /role not set/) {
		# cvm is not active
		VCSAG_LOG_MSG("E", "vxconfigd is not in the correct clustered state: $line",
			54, "$line");
		return -1;
	}
	if ($fields[5] eq "MASTER") {
		# This node is the master.
		return 1;
	}
	elsif ($fields[5] ne "MASTER") {
		# This node is a slave.
		return 0;
	}
}

1;
